import numpy as np
import pandas as pd
from spinesTS.preprocessing import split_series
from spinesUtils import ParameterTypeAssert, ParameterValuesAssert

from PipelineTS.base import NNModelMixin, IntervalEstimationMixin


class SpinesNNModelMixin(NNModelMixin, IntervalEstimationMixin):

    def __init__(self, time_col, target_col, device=None):

        super().__init__(time_col, target_col, device=device)
        self.last_x = None

    def _define_model(self):
        raise NotImplementedError

    @ParameterValuesAssert({
        'mode': ('train', 'predict')
    })
    def _data_preprocess(self, data, update_last_data=False, mode='train'):
        data[self.all_configs['time_col']] = pd.to_datetime(data[self.all_configs['time_col']])
        if update_last_data:
            self.last_dt = data[self.all_configs['time_col']].max()

        if mode == 'train':
            if update_last_data:
                self.last_x = data[self.all_configs['target_col']].iloc[-(2 * self.all_configs['lags']):]
            # x_train, y_train
            x_train, y_train = split_series(data[self.all_configs['target_col']], data[self.all_configs['target_col']],
                                            window_size=self.all_configs['lags'], pred_steps=self.all_configs['lags'])

            if x_train.ndim == 1:
                x_train = x_train.reshape(1, -1)

            if y_train.ndim == 1:
                y_train = y_train.reshape(1, -1)

            return x_train, y_train
        else:
            x, y = split_series(pd.concat((self.last_x, data[self.all_configs['target_col']])),
                                pd.concat((self.last_x, data[self.all_configs['target_col']])),
                                window_size=self.all_configs['lags'], pred_steps=self.all_configs['lags'])

            if x.ndim == 1:
                x = x.reshape(1, -1)

            if y.ndim == 1:
                y = y.reshape(1, -1)

            return x, y

    @ParameterTypeAssert({
        'valid_data': (None, pd.DataFrame)
    })
    def fit(self, data, valid_data=None, cv=5, fit_kwargs=None):
        data = data[[self.all_configs['time_col'], self.all_configs['target_col']]]

        if fit_kwargs is None:
            fit_kwargs = {}

        for fit_param in [
            'verbose', 'epochs', 'batch_size', 'patience',
            'min_delta', 'lr_scheduler', 'lr_scheduler_patience',
            'lr_factor', 'restore_best_weights'
        ]:
            if fit_param not in fit_kwargs:
                fit_kwargs.update({fit_param: self.all_configs[fit_param]})

        x, y = self._data_preprocess(data, update_last_data=True, mode='train')

        self.x = data[self.all_configs['target_col']].iloc[-self.all_configs['lags']:]

        if valid_data is None:
            eval_set = [(x, y)]
        else:
            valid_x, valid_y = self._data_preprocess(valid_data, update_last_data=False, mode='predict')

            eval_set = [(valid_x, valid_y)]

        self.model.fit(x, y, eval_set=eval_set, **fit_kwargs)

        if self.all_configs['quantile'] is not None:
            self.all_configs['quantile_error'] = \
                self.calculate_confidence_interval_nn(data, fit_kwargs=fit_kwargs, cv=cv)

        return self

    def _extend_predict(self, x, n, predict_kwargs):
        """Extrapolation prediction.

        Parameters
        ----------
        x: to_predict data, must be 2 dims data
        n: predict steps, must be int

        Returns
        -------
        np.ndarray, which has 2 dims

        """

        assert isinstance(n, int)
        assert x.ndim == 2

        current_res = self.model.predict(x, **predict_kwargs)

        if n is None:
            return current_res.squeeze().tolist()
        elif n <= current_res.shape[1]:
            return current_res[-1][:n].tolist()
        else:
            res = current_res.squeeze().tolist()
            for i in range(n - self.all_configs['lags']):
                x = np.concatenate((x[:, 1:], current_res[:, 0:1]), axis=1)
                current_res = self.model.predict(x, **predict_kwargs)

                res.append(current_res.squeeze().tolist()[-1])

            return res

    def predict(self, n, predict_kwargs=None):
        if predict_kwargs is None:
            predict_kwargs = {}

        x = self.x.values.reshape(1, -1)
        res = self._extend_predict(x, n, predict_kwargs=predict_kwargs)  # list

        assert len(res) == n

        res = pd.DataFrame(res, columns=[self.all_configs['target_col']])
        res[self.all_configs['time_col']] = \
            self.last_dt + pd.to_timedelta(range(res.index.shape[0] + 1), unit='D')[1:]

        if self.all_configs['quantile'] is not None:
            res = self.interval_predict(res)

        return self.chosen_cols(res)



from spinesTS.ml_model import MultiOutputRegressor as MOR, MultiStepRegressor as MSR

import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from lightgbm import LGBMRegressor

from spinesTS.pipeline import Pipeline
from spinesTS.preprocessing import split_series
from spinesUtils import generate_function_kwargs

from PipelineTS.base import GBDTModelMixin, IntervalEstimationMixin


class MultiOutputRegressorModel(GBDTModelMixin, IntervalEstimationMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            quantile=0.9,
            estimator=LGBMRegressor,
            use_standard_scaler=None,
            **model_configs
    ):
        super().__init__(time_col=time_col, target_col=target_col)

        self._base_estimator = estimator
        self.all_configs['model_configs'] = generate_function_kwargs(
            self._base_estimator,
            **model_configs
        )

        self.last_dt = None
        self.last_lags_dataframe = None

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0,
                'use_standard_scaler': use_standard_scaler
            }
        )

        self.x = None

        self.model = self._define_model()

    def _define_model(self):
        if self.all_configs['use_standard_scaler'] is not None:
            if self.all_configs['use_standard_scaler']:
                model = Pipeline([
                    ('scaler', StandardScaler()),
                    ('estimator', MOR(
                        self._base_estimator(**self.all_configs['model_configs'])
                    ))
                ])
            else:
                model = Pipeline([
                    ('scaler', MinMaxScaler()),
                    ('estimator', MOR(
                        self._base_estimator(**self.all_configs['model_configs'])
                    ))
                ])
        else:
            model = Pipeline([
                ('estimator', MOR(
                    self._base_estimator(**self.all_configs['model_configs'])
                ))
            ])

        return model

    def _data_preprocess(self, data, update_last_dt=False):
        data[self.all_configs['time_col']] = pd.to_datetime(data[self.all_configs['time_col']])
        if update_last_dt:
            self.last_dt = data[self.all_configs['time_col']].max()

        # x_train, y_train
        return split_series(data[self.all_configs['target_col']], data[self.all_configs['target_col']],
                            window_size=self.all_configs['lags'], pred_steps=self.all_configs['lags'])

    def fit(self, data, cv=5, fit_kwargs=None):
        data = data[[self.all_configs['time_col'], self.all_configs['target_col']]]

        self.last_lags_dataframe = data.iloc[-(2 * self.all_configs['lags'] + 1):, :]

        if fit_kwargs is None:
            fit_kwargs = {}

        x, y = self._data_preprocess(data, update_last_dt=True)
        x = pd.DataFrame(x)

        self.x = x.iloc[-1:, :]

        self.model.fit(x, y, **fit_kwargs)

        if self.all_configs['quantile'] is not None:
            self.all_configs['quantile_error'] = \
                self.calculate_confidence_interval_mor(data, fit_kwargs=fit_kwargs, cv=cv)

        return self

    def _extend_predict(self, x, n, predict_kwargs):
        """Extrapolation prediction.

        Parameters
        ----------
        x: to_predict data, must be 2 dims data
        n: predict steps, must be int

        Returns
        -------
        np.ndarray, which has 2 dims

        """

        assert isinstance(n, int)
        assert x.ndim == 2

        current_res = self.model.predict(x, **predict_kwargs)

        if n is None:
            return current_res.squeeze().tolist()
        elif n <= current_res.shape[1]:
            return current_res.squeeze().tolist()[:n]
        else:
            res = current_res.squeeze().tolist()
            for i in range(n - self.all_configs['lags']):
                x = np.concatenate((x[:, 1:], current_res[:, 0:1]), axis=1)
                current_res = self.model.predict(x, **predict_kwargs)

                res.append(current_res.squeeze().tolist()[-1])

            return res

    def predict(self, n, predict_kwargs=None):
        if predict_kwargs is None:
            predict_kwargs = {}

        x = self.x.values.reshape(1, -1)
        res = self._extend_predict(x, n, predict_kwargs=predict_kwargs)  # list

        assert len(res) == n

        res = pd.DataFrame(res, columns=[self.all_configs['target_col']])
        res[self.all_configs['time_col']] = \
            self.last_dt + pd.to_timedelta(range(res.index.shape[0] + 1), unit='D')[1:]

        if self.all_configs['quantile'] is not None:
            res = self.interval_predict(res)

        return self.chosen_cols(res)


class MultiStepRegressorModel(GBDTModelMixin, IntervalEstimationMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            quantile=0.9,
            estimator=LGBMRegressor,
            use_standard_scaler=None,
            **model_configs
    ):
        super().__init__(time_col=time_col, target_col=target_col)

        self._base_estimator = estimator
        self.all_configs['model_configs'] = generate_function_kwargs(
            self._base_estimator,
            **model_configs
        )

        self.last_dt = None
        self.last_lags_dataframe = None

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0,
                'use_standard_scaler': use_standard_scaler
            }
        )

        self.x = None

        self.model = self._define_model()

    def _define_model(self):
        if self.all_configs['use_standard_scaler'] is not None:
            if self.all_configs['use_standard_scaler']:
                model = Pipeline([
                    ('scaler', StandardScaler()),
                    ('estimator', MSR(
                        self._base_estimator(**self.all_configs['model_configs'])
                    ))
                ])
            else:
                model = Pipeline([
                    ('scaler', MinMaxScaler()),
                    ('estimator', MSR(
                        self._base_estimator(**self.all_configs['model_configs'])
                    ))
                ])
        else:
            model = Pipeline([
                ('estimator', MSR(
                    self._base_estimator(**self.all_configs['model_configs'])
                ))
            ])

        return model

    def _data_preprocess(self, data, update_last_dt=False):
        data[self.all_configs['time_col']] = pd.to_datetime(data[self.all_configs['time_col']])
        if update_last_dt:
            self.last_dt = data[self.all_configs['time_col']].max()

        # x_train, y_train
        return split_series(data[self.all_configs['target_col']], data[self.all_configs['target_col']],
                            window_size=self.all_configs['lags'], pred_steps=self.all_configs['lags'])

    def fit(self, data, cv=5, fit_kwargs=None):
        data = data[[self.all_configs['time_col'], self.all_configs['target_col']]]

        self.last_lags_dataframe = data.iloc[-(2 * self.all_configs['lags'] + 1):, :]

        if fit_kwargs is None:
            fit_kwargs = {}

        x, y = self._data_preprocess(data, update_last_dt=True)
        x = pd.DataFrame(x)

        self.x = x.iloc[-1:, :]

        self.model.fit(x, y, **fit_kwargs)

        if self.all_configs['quantile'] is not None:
            self.all_configs['quantile_error'] = \
                self.calculate_confidence_interval_mor(data, fit_kwargs=fit_kwargs, cv=cv)

        return self

    def _extend_predict(self, x, n, predict_kwargs):
        """Extrapolation prediction.

        Parameters
        ----------
        x: to_predict data, must be 2 dims data
        n: predict steps, must be int

        Returns
        -------
        np.ndarray, which has 2 dims

        """

        assert isinstance(n, int)
        assert x.ndim == 2

        current_res = self.model.predict(x, **predict_kwargs).reshape(1, -1)

        if n is None:
            return current_res.squeeze().tolist()
        elif n <= current_res.shape[1]:
            return current_res.squeeze().tolist()[:n]
        else:
            res = current_res.squeeze().tolist()
            for i in range(n - self.all_configs['lags']):
                x = np.concatenate((x[:, 1:], current_res[:, 0:1]), axis=1)
                current_res = self.model.predict(x, **predict_kwargs).reshape(1, -1)

                res.append(current_res.squeeze().tolist()[-1])

            return res

    def predict(self, n, predict_kwargs=None):
        if predict_kwargs is None:
            predict_kwargs = {}

        x = self.x.values.reshape(1, -1)
        res = self._extend_predict(x, n, predict_kwargs=predict_kwargs)  # list

        assert len(res) == n

        res = pd.DataFrame(res, columns=[self.all_configs['target_col']])
        res[self.all_configs['time_col']] = \
            self.last_dt + pd.to_timedelta(range(res.index.shape[0] + 1), unit='D')[1:]

        if self.all_configs['quantile'] is not None:
            res = self.interval_predict(res)

        return self.chosen_cols(res)


import numpy as np

from spinesUtils.asserts import ParameterTypeAssert


@ParameterTypeAssert({
    'yt': np.ndarray,
    'left_pred': np.ndarray,
    'right_pred': np.ndarray
})
def quantile_acc(yt, left_pred, right_pred):
    """计算预测区间准确率"""
    return np.sum((yt >= left_pred) * (yt <= right_pred)) / yt.shape[0]



from copy import deepcopy

import pandas as pd
import numpy as np

from spinesTS.nn import StackingRNN
from spinesTS.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from spinesUtils import generate_function_kwargs

from PipelineTS.nn_model.sps_nn_model_base import SpinesNNModelMixin


class StackingRNNModel(SpinesNNModelMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=30,
            quantile=0.9,
            random_state=None,
            dropout=0.1,
            learning_rate=0.001,
            device='cpu',
            use_standard_scaler=False,
            verbose=False,
            epochs=1000,
            batch_size='auto',
            patience=100,
            min_delta=0,
            lr_scheduler='ReduceLROnPlateau',
            lr_scheduler_patience=10,
            lr_factor=0.7,
            restore_best_weights=True,
    ):
        super().__init__(time_col=time_col, target_col=target_col, device=device)

        self.all_configs['model_configs'] = generate_function_kwargs(
            StackingRNN,
            in_features=lags,
            out_features=lags,
            loss_fn='mae',
            bias=True,
            dropout=dropout,
            learning_rate=learning_rate,
            random_seed=random_state,
            device=self.device,
        )

        self.last_dt = None

        self.all_configs.update(
            {
                'lags': lags,
                'quantile': quantile,
                'time_col': time_col,
                'target_col': target_col,
                'quantile_error': 0,
                'use_standard_scaler': use_standard_scaler,
                'verbose': verbose,
                'epochs': epochs,
                'batch_size': batch_size,
                'patience': patience,
                'min_delta': min_delta,
                'lr_scheduler': lr_scheduler,
                'lr_scheduler_patience': lr_scheduler_patience,
                'lr_factor': lr_factor,
                'restore_best_weights': restore_best_weights
            }
        )

        self.x = None

        self.model = self._define_model()

    def _define_model(self):
        if self.all_configs['use_standard_scaler'] is not None:
            if self.all_configs['use_standard_scaler']:
                model = Pipeline([
                    ('scaler', StandardScaler()),
                    ('estimator', StackingRNN(**self.all_configs['model_configs']))
                ])
            else:
                model = Pipeline([
                    ('scaler', MinMaxScaler()),
                    ('estimator', StackingRNN(**self.all_configs['model_configs']))
                ])
        else:
            model = StackingRNN(**self.all_configs['model_configs'])

        return model

    def _extend_predict(self, x, n, predict_kwargs):
        """Extrapolation prediction.

        Parameters
        ----------
        x: to_predict data, must be 2 dims data
        n: predict steps, must be int

        Returns
        -------
        np.ndarray, which has 2 dims

        """

        assert isinstance(n, int)
        assert x.ndim == 2

        current_res = self.model.predict(x, **predict_kwargs)

        if n is None:
            return current_res[0].squeeze().tolist()
        elif n <= current_res.shape[1]:
            return current_res[0].squeeze()[:n].tolist()
        else:
            res = deepcopy(current_res)

            for i in range(n - self.all_configs['lags']):
                x = np.concatenate((x[:, 1:], current_res[:, 0:1]), axis=-1)
                current_res = self.model.predict(x, **predict_kwargs)  # 2D

                res = np.concatenate((res, current_res[:, -1:]), axis=-1)

            return res[0].squeeze().tolist()

    def predict(self, n, predict_kwargs=None):
        if predict_kwargs is None:
            predict_kwargs = {}

        x = self.x.values.reshape(1, -1)
        x = np.concatenate((x, x), axis=0)

        res = self._extend_predict(x, n, predict_kwargs=predict_kwargs)  # list

        assert len(res) == n

        res = pd.DataFrame(res, columns=[self.all_configs['target_col']])
        res[self.all_configs['time_col']] = \
            self.last_dt + pd.to_timedelta(range(res.index.shape[0] + 1), unit='D')[1:]

        if self.all_configs['quantile'] is not None:
            res = self.interval_predict(res)

        return self.chosen_cols(res)




import sys
from copy import deepcopy

import numpy as np
from spinesUtils.asserts import *
from spinesTS.metrics import wmape
from spinesTS.utils import func_has_params


def load_dataset_to_darts(
        data,
        time_col,
        target_col,
        fill_missing_dates=False,
        freq=None,
        fillna_value=None,
        static_covariates=None,
        hierarchy=None
):
    from darts.timeseries import TimeSeries

    return TimeSeries.from_dataframe(
        data,
        time_col=time_col,
        value_cols=target_col,
        fill_missing_dates=fill_missing_dates,
        freq=freq,
        fillna_value=fillna_value,
        static_covariates=static_covariates,
        hierarchy=hierarchy
    )


class DartsForecastMixin:
    @staticmethod
    def convert2dts_dataframe(
            df,
            time_col,
            target_col,
            **kwargs
    ):
        return load_dataset_to_darts(df, time_col, target_col, **kwargs)

    @staticmethod
    def convert2pd_dataframe(df):
        return df.pd_dataframe()

    def fit(self, data, convert_dataframe_kwargs=None, fit_kwargs=None, convert_float32=True):
        if convert_dataframe_kwargs is None:
            convert_dataframe_kwargs = {}
        if fit_kwargs is None:
            fit_kwargs = {}

        data = self.convert2dts_dataframe(data, time_col=self.all_configs['time_col'],
                                          target_col=self.all_configs['target_col'],
                                          **convert_dataframe_kwargs)

        if convert_float32:
            data = data.astype(np.float32)

        self.model.fit(data, **fit_kwargs)
        return self

    def predict(self, n, predict_likelihood_parameters=False, predict_kwargs=None):
        if predict_kwargs is None:
            predict_kwargs = {}
        if 'predict_likelihood_parameters' in get_function_params_name(self.model.predict):
            return self.model.predict(
                n,
                predict_likelihood_parameters=predict_likelihood_parameters,
                **predict_kwargs
            ).pd_dataframe()
        return self.model.predict(n, **predict_kwargs).pd_dataframe()

    def rename_prediction(self, data):
        data.columns.name = None
        data[self.all_configs['time_col']] = data.index.copy()

        data = data.reset_index(drop=True)

        for i in data.columns:
            if i == f"{self.all_configs['target_col']}_q0.50":
                data.rename(columns={i: f"{self.all_configs['target_col']}"}, inplace=True)

        if self.all_configs['quantile'] is not None:
            if self.all_configs['quantile'] < round(1 - self.all_configs['quantile'], 1):
                ratio = self.all_configs['quantile']
            else:
                ratio = round(1 - self.all_configs['quantile'], 1)

            if len(str(ratio).split('.')[-1]) == 1:
                left_ratio = str(ratio) + '0'
                right_ratio = str(1 - ratio) + '0'
            else:
                left_ratio = str(ratio)
                right_ratio = str(1 - ratio)

            for i in data.columns:
                if i == f"{self.all_configs['target_col']}_q{right_ratio}":
                    data.rename(columns={i: f"{self.all_configs['target_col']}_upper"}, inplace=True)

                elif i == f"{self.all_configs['target_col']}_q{left_ratio}":
                    data.rename(columns={i: f"{self.all_configs['target_col']}_lower"}, inplace=True)

        return self.chosen_cols(data)


class GBDTModelMixin:
    def __init__(self, time_col, target_col):
        self.all_configs = {'model_configs': {}}
        self.sorted_cols = [
            time_col,
            target_col,
            f"{target_col}_lower",
            f"{target_col}_upper"
        ]

    def chosen_cols(self, data):
        if all(i in data.columns for i in self.sorted_cols):
            return data[self.sorted_cols]
        else:
            return data[[self.all_configs['time_col'],
                         *[i for i in data.columns if i != self.all_configs['time_col']]]]


class StatisticModelMixin:
    def __init__(self, time_col, target_col):
        self.all_configs = {'model_configs': {}}
        self.sorted_cols = [
            time_col,
            target_col,
            f"{target_col}_lower",
            f"{target_col}_upper"
        ]

    def chosen_cols(self, data):
        if all(i in data.columns for i in self.sorted_cols):
            return data[self.sorted_cols]
        else:
            return data[[self.all_configs['time_col'],
                         *[i for i in data.columns if i != self.all_configs['time_col']]]]


class NNModelMixin:
    def __init__(self, time_col, target_col, device=None):
        self.all_configs = {'model_configs': {}}
        if device is None:
            if sys.platform == 'darwin':
                self.device = 'cpu'
            else:
                self.device = 'auto'
        else:
            self.device = device

        self.sorted_cols = [
            time_col,
            target_col,
            f"{target_col}_lower",
            f"{target_col}_upper"
        ]

    def chosen_cols(self, data):
        if all(i in data.columns for i in self.sorted_cols):
            return data[self.sorted_cols]
        else:
            return data[[self.all_configs['time_col'],
                         *[i for i in data.columns if i != self.all_configs['time_col']]]]


class IntervalEstimationMixin:
    def check_data(self, data):
        from darts.timeseries import TimeSeries
        if isinstance(data, TimeSeries):
            data = data.pd_dataframe()

        if len(data) < 2 * self.all_configs['lags']:
            raise ValueError("data length must be greater than or equal to 2 * lags.")

    def _split_train_valid_data(self, data, cv=5, is_prophet=False):
        self.check_data(data)

        if is_prophet:
            data = data[['ds', 'y']]
        else:
            data = data[[self.all_configs['time_col'], self.all_configs['target_col']]]

        from mapie.subsample import BlockBootstrap

        cv = BlockBootstrap(n_resamplings=cv, length=self.all_configs['lags'], random_state=0)

        for train_index, test_index in cv.split(data):
            if len(test_index) > 0 and len(train_index) >= self.all_configs['lags']:
                yield (data.iloc[train_index, :].reset_index(drop=True),
                       data.iloc[test_index, :].reset_index(drop=True))

    def calculate_confidence_interval_darts(self, data, cv=5, fit_kwargs=None, convert2dts_dataframe_kwargs=None):
        if fit_kwargs is None:
            fit_kwargs = {}

        if convert2dts_dataframe_kwargs is None:
            convert2dts_dataframe_kwargs = {}

        residuals = []

        for train_data, valid_data in self._split_train_valid_data(data, cv=cv):
            valid_y = valid_data[[self.all_configs['target_col']]].values

            train_ds = self.convert2dts_dataframe(
                data.reset_index(drop=True),
                time_col=self.all_configs['time_col'],
                target_col=self.all_configs['target_col'],
                **convert2dts_dataframe_kwargs
            ).astype(np.float32)

            model = self._define_model()

            model.fit(train_ds, **fit_kwargs)

            res = model.predict(len(valid_y)).pd_dataframe()[self.all_configs['target_col']].values

            y_cal_error = wmape(valid_y.flatten(), res.flatten())

            residuals.append(y_cal_error)

        quantile = np.percentile(residuals, q=self.all_configs['quantile'])

        return quantile

    def _calculate_confidence_interval_sps(self, data, cv=5, fit_kwargs=None, train_data_process_kwargs=None,
                                           valid_data_process_kwargs=None):
        if fit_kwargs is None:
            fit_kwargs = {}

        if train_data_process_kwargs is None:
            train_data_process_kwargs = {}

        if valid_data_process_kwargs is None:
            valid_data_process_kwargs = {}

        residuals = []
        for train_data, valid_data in self._split_train_valid_data(data, cv=cv):

            data_x, data_y = self._data_preprocess(train_data, **train_data_process_kwargs)

            valid_data_x, valid_data_y = self._data_preprocess(valid_data, **valid_data_process_kwargs)

            model = self._define_model()

            if func_has_params(model.fit, 'eval_set'):
                model.fit(data_x, data_y, eval_set=[(data_x, data_y)], **fit_kwargs)
            else:
                model.fit(data_x, data_y, **fit_kwargs)

            res = model.predict(valid_data_x).flatten()

            y_cal_error = wmape(valid_data_y.flatten(), res.flatten())

            residuals.append(y_cal_error)

        quantile = np.percentile(y_cal_error, q=self.all_configs['quantile'])

        return quantile

    def calculate_confidence_interval_mor(self, data, cv=5, fit_kwargs=None):
        return self._calculate_confidence_interval_sps(data, fit_kwargs=fit_kwargs, cv=cv)

    def calculate_confidence_interval_gbrt(self, data, cv=5, fit_kwargs=None):

        return self._calculate_confidence_interval_sps(data, fit_kwargs=fit_kwargs,
                                                       train_data_process_kwargs={'mode': 'train'},
                                                       valid_data_process_kwargs={'mode': 'train'},
                                                       cv=cv)

    def calculate_confidence_interval_nn(self, data, cv=5, fit_kwargs=None):
        if fit_kwargs is None:
            kwargs = {}
        else:
            kwargs = deepcopy(fit_kwargs)

        kwargs.update({'verbose': False})

        return self._calculate_confidence_interval_sps(
            data, fit_kwargs=kwargs, train_data_process_kwargs={'mode': 'train', 'update_last_data': False},
            valid_data_process_kwargs={'mode': 'train', 'update_last_data': False}, cv=cv)

    def calculate_confidence_interval_prophet(self, data, cv=5, freq='D', fit_kwargs=None):
        if fit_kwargs is None:
            fit_kwargs = {}

        for train_data, valid_data in self._split_train_valid_data(data, cv=cv, is_prophet=True):
            train_ds = train_data[['ds', 'y']]

            residuals = []

            valid_data_y = valid_data['y'].values

            model = self._define_model()

            model.fit(train_ds, **fit_kwargs)

            res = model.predict(
                self.model.make_future_dataframe(
                    periods=len(valid_data_y),
                    freq=freq,
                    include_history=False,
                ))['yhat'].values

            y_cal_error = wmape(valid_data_y.flatten(), res.flatten())

            residuals.append(y_cal_error)

        quantile = np.percentile(residuals, q=self.all_configs['quantile'])

        return quantile

    def interval_predict(self, res):
        res[f"{self.all_configs['target_col']}_lower"] = \
            res[self.all_configs['target_col']].values * (1 - self.all_configs['quantile_error'])
        res[f"{self.all_configs['target_col']}_upper"] = \
            res[self.all_configs['target_col']].values * (1 + self.all_configs['quantile_error'])

        return self.chosen_cols(res)



import logging

from prophet import Prophet
from spinesUtils import generate_function_kwargs

from PipelineTS.base import StatisticModelMixin, IntervalEstimationMixin

logger = logging.getLogger('cmdstanpy')
logger.addHandler(logging.NullHandler())
logger.propagate = False
logger.setLevel(logging.CRITICAL)


class ProphetModel(StatisticModelMixin, IntervalEstimationMixin):
    def __init__(
            self,
            time_col,
            target_col,
            lags=1,
            country_holidays=None,
            quantile=0.9,
            random_state=0,
            **prophet_configs
    ):
        super().__init__(time_col=time_col, target_col=target_col)

        self.all_configs['model_configs'] = generate_function_kwargs(
            Prophet,
            holidays=country_holidays,
            **prophet_configs
        )

        self.model = self._define_model()

        self.all_configs.update({
            'quantile': quantile,
            'quantile_error': 0,
            'time_col': time_col,
            'target_col': target_col,
            'random_state': random_state,  # meanness, but only to follow coding conventions
            'lags': lags,  # meanness, but only to follow coding conventions
        })

    def _define_model(self):
        return Prophet(**self.all_configs['model_configs'])

    @staticmethod
    def _prophet_preprocessing(df, time_col, target_col):
        df_ = df[[time_col, target_col]]
        if 'ds' != time_col or 'y' != target_col:
            df_ = df_.rename(columns={time_col: 'ds', target_col: 'y'})

        return df_

    def fit(self, data, freq='D', cv=5, fit_kwargs=None):
        if fit_kwargs is None:
            fit_kwargs = {}
        data = self._prophet_preprocessing(data, self.all_configs['time_col'], self.all_configs['target_col'])
        self.model.fit(data, **fit_kwargs)

        if self.all_configs['quantile'] is not None:
            self.all_configs['quantile_error'] = \
                self.calculate_confidence_interval_prophet(data, cv=cv, freq=freq, fit_kwargs=fit_kwargs)
        return self

    def predict(self, num_days, freq='D', include_history=False):
        res = self.model.predict(
            self.model.make_future_dataframe(
                periods=num_days,
                freq=freq,
                include_history=include_history,
            )
        )[['ds', 'yhat']].rename(
            columns={
                'ds': self.all_configs['time_col'],
                'yhat': self.all_configs['target_col'],
            }
        )

        if self.all_configs['quantile'] is not None:
            res = self.interval_predict(res)

        return self.chosen_cols(res)



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='PipelineTS',
    version="0.3.3",
    description='One-stop time series analysis tool, supporting time series data preprocessing, '
                'feature engineering, model training, model evaluation, and model prediction.',
    keywords='timeseries machine learning',
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12'
    ],
    url='https://github.com/BirchKwok/PipelineTS',
    author='Birch Kwok',
    author_email='birchkwok@gmail.com',
    install_requires=[
        'scikit-learn>=1.3.0',
        'numpy>=1.24.3',
        'pandas>=2.0.3',
        'matplotlib>=3.7.1',
        'frozendict>=2.3.0',
        'darts>=0.24.0',
        'prophet>=1.1.4',
        'spinesTS>=0.3.12',
        'spinesUtils>=0.3.5',
        'lightgbm>=3.3.5',
        'IPython>=8.12.1',
        'tabulate>=0.8.9',
        'torch>=2.0.0',
        'xgboost>=2.0.0',
        'catboost>=1.2.2',
        'MAPIE>=0.7.0'
    ],
    zip_safe=False,
    include_package_data=True
)



import pandas as pd

from spinesUtils.asserts import ParameterTypeAssert, ParameterValuesAssert


@ParameterTypeAssert({
    'data1': pd.DataFrame,
    'data2': pd.DataFrame,
    'time_col': str,
    'target_col': str,
    'labels': (None, list, tuple)
})
def plot_data_period(data1, data2, time_col, target_col, labels=None):
    """可视化日期数据"""
    import matplotlib.pyplot as plt
    import matplotlib.dates as m_dates

    assert data1[time_col].iloc[0] <= data2[time_col].iloc[0], \
        'The starting time of data1 must be before data2, or equal to the starting time of data2.'

    if f'{target_col}_upper' in data2.columns and f'{target_col}_lower' in data2.columns:
        if labels is None:
            labels = ('True Data', 'Prediction Data', 'Upper Bound', 'Lower Bound')
        else:
            assert len(labels) == 4
    else:
        if labels is None:
            labels = ('Data', 'Test Data')
        else:
            assert len(labels) == 2

    # 绘制曲线
    plt.plot_date(data1[time_col], data1[target_col], label=labels[0],
                  color='black', linestyle='-', marker='')
    plt.plot_date(data2[time_col], data2[target_col],
                  label=labels[1], color='blue', linestyle='-', marker='')

    if f'{target_col}_upper' in data2.columns and  f'{target_col}_lower' in data2.columns:
        plt.plot_date(data2[time_col], data2[f'{target_col}_upper'],
                      label=labels[2], color='orchid', linestyle='-', marker='')
        plt.plot_date(data2[time_col], data2[f'{target_col}_lower'],
                      label=labels[3], color='violet', linestyle='-', marker='')
        plt.fill_between(data2[time_col], data2[f'{target_col}_upper'], data2[f'{target_col}_lower'],
                         color='dimgray', alpha=0.2)

    if data2[time_col].iloc[0] > data1[time_col].iloc[0]:
        # 添加竖线
        plt.axvline(x=data2[time_col].iloc[0], color='black', linestyle='--')

    valid_data_max = data2[target_col].max()
    valid_period = [data2[time_col].iloc[0], data2[time_col].iloc[-1]]

    plt.axvspan(valid_period[0], valid_period[1], ymin=0, ymax=valid_data_max,
                facecolor='lightyellow', alpha=0.5)

    # 设置x轴为日期格式
    date_fmt = m_dates.DateFormatter('%Y-%m-%d')
    plt.gca().xaxis.set_major_formatter(date_fmt)

    # 添加网格线
    plt.grid(True)

    # 取消外框
    plt.box(False)

    # 自动调整日期标签的格式和位置，以避免重叠
    plt.gcf().autofmt_xdate()
    # 添加图例
    plt.legend()

    # 显示图形
    plt.show()
